/**
 * Copyright (C) 2019 Bonitasoft S.A.
 * Bonitasoft, 32 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301, USA.
 **/
package org.bonitasoft.engine.business.application;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.fail;

import org.bonitasoft.engine.search.Order;
import org.bonitasoft.engine.search.SearchOptions;
import org.bonitasoft.engine.search.SearchOptionsBuilder;
import org.bonitasoft.engine.search.SearchResult;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

/**
 * @author Elias Ricken de Medeiros
 */
public class LivingApplicationMenuIT extends TestWithCustomPage {

    private Application application;

    private ApplicationPage appPage;

    @Override
    @Before
    public void setUp() throws Exception {
        super.setUp();
        application = getLivingApplicationAPI().createApplication(new ApplicationCreator("app", "My app", "1.0"));
        appPage = getLivingApplicationAPI().createApplicationPage(application.getId(), getPage().getId(), "myPage");
    }

    @Override
    @After
    public void tearDown() throws Exception {
        getLivingApplicationAPI().deleteApplication(application.getId());
        application = null;
        appPage = null;
        super.tearDown();
    }

    @Test
    public void createApplicationMenu_ApplicationPage_should_return_applicationMenu_based_on_creator_and_should_manage_indexes()
            throws Exception {
        //given
        final ApplicationMenuCreator creator = new ApplicationMenuCreator(application.getId(), "Main");

        //when
        final ApplicationMenu createdAppMenu = getLivingApplicationAPI().createApplicationMenu(creator);

        //then
        assertThat(createdAppMenu).isNotNull();
        assertThat(createdAppMenu.getDisplayName()).isEqualTo("Main");
        assertThat(createdAppMenu.getApplicationId()).isEqualTo(application.getId());
        assertThat(createdAppMenu.getApplicationPageId()).isNull();
        assertThat(createdAppMenu.getIndex()).isEqualTo(1);
        assertThat(createdAppMenu.getParentId()).isNull();
        assertThat(createdAppMenu.getId()).isGreaterThan(0);

        //when
        //create a second menu
        ApplicationMenu index2Menu = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second menu"));

        //then
        // should have a index incremented
        assertThat(index2Menu.getIndex()).isEqualTo(2);

    }

    @Test
    public void createApplicationMenu_with_applicationPage_should_return_ApplicationMenu_with_applicationPageId()
            throws Exception {
        //given
        final ApplicationMenuCreator creator = new ApplicationMenuCreator(application.getId(), "Main", appPage.getId());

        //when
        final ApplicationMenu createdAppMenu = getLivingApplicationAPI().createApplicationMenu(creator);

        //then
        assertThat(createdAppMenu).isNotNull();
        assertThat(createdAppMenu.getApplicationPageId()).isEqualTo(appPage.getId());
    }

    @Test
    public void createApplicationMenu_with_parent_menu_should_return_ApplicationMenu_with_parentId_and_should_manage_indexes()
            throws Exception {
        //given
        final ApplicationMenuCreator mainCreator = new ApplicationMenuCreator(application.getId(), "Main");
        final ApplicationMenu mainMenu = getLivingApplicationAPI().createApplicationMenu(mainCreator);

        final ApplicationMenuCreator childCreator = new ApplicationMenuCreator(application.getId(), "Child",
                appPage.getId());
        childCreator.setParentId(mainMenu.getId());

        //when
        final ApplicationMenu createdAppMenu = getLivingApplicationAPI().createApplicationMenu(childCreator);

        //then
        assertThat(createdAppMenu).isNotNull();
        assertThat(createdAppMenu.getParentId()).isEqualTo(mainMenu.getId());
        assertThat(createdAppMenu.getIndex()).isEqualTo(1);

        //when
        //create a child menu second menu
        ApplicationMenuCreator secondChildCreator = new ApplicationMenuCreator(application.getId(), "second child",
                appPage.getId());
        secondChildCreator.setParentId(mainMenu.getId());
        ApplicationMenu secondChild = getLivingApplicationAPI().createApplicationMenu(secondChildCreator);

        //then
        //should have incremented index
        assertThat(secondChild.getIndex()).isEqualTo(2);

    }

    @Test
    public void updateApplicationMenu_should_update_application_menu_based_on_updater() throws Exception {
        //given
        final ApplicationMenuCreator parentCreator = new ApplicationMenuCreator(application.getId(), "Main");
        final ApplicationMenuCreator childCreator = new ApplicationMenuCreator(application.getId(), "Child");
        final ApplicationMenu parentAppMenu = getLivingApplicationAPI().createApplicationMenu(parentCreator);
        final ApplicationMenu childCreatedAppMenu = getLivingApplicationAPI().createApplicationMenu(childCreator);

        ApplicationMenuUpdater updater = new ApplicationMenuUpdater();
        updater.setApplicationPageId(appPage.getId());
        updater.setParentId(parentAppMenu.getId());
        updater.setDisplayName("Updated child");

        //when
        ApplicationMenu updatedChildMenu = getLivingApplicationAPI().updateApplicationMenu(childCreatedAppMenu.getId(),
                updater);

        //then
        assertThat(updatedChildMenu).isNotNull();
        //updated:
        assertThat(updatedChildMenu.getDisplayName()).isEqualTo("Updated child");
        assertThat(updatedChildMenu.getApplicationPageId()).isEqualTo(appPage.getId());
        assertThat(updatedChildMenu.getParentId()).isEqualTo(parentAppMenu.getId());
        assertThat(updatedChildMenu.getIndex()).isEqualTo(1); //because parent changed
        //not changed:
        assertThat(updatedChildMenu.getApplicationId()).isEqualTo(application.getId());

        //given
        //check it's possible to clean parent and application page
        updater = new ApplicationMenuUpdater();
        updater.setApplicationPageId(null);
        updater.setParentId(null);

        //when
        updatedChildMenu = getLivingApplicationAPI().updateApplicationMenu(childCreatedAppMenu.getId(), updater);

        //then
        assertThat(updatedChildMenu).isNotNull();
        assertThat(updatedChildMenu).isEqualTo(getLivingApplicationAPI().getApplicationMenu(updatedChildMenu.getId()));
        // updated:
        assertThat(updatedChildMenu.getApplicationPageId()).isNull();
        assertThat(updatedChildMenu.getParentId()).isNull();
        assertThat(updatedChildMenu.getIndex()).isEqualTo(2); //because parent changed
        //not changed:
        assertThat(updatedChildMenu.getDisplayName()).isEqualTo("Updated child");
        assertThat(updatedChildMenu.getApplicationId()).isEqualTo(application.getId());

    }

    @Test
    public void updateApplicationMenu_index_should_organize_indexes_for_elements_having_the_same_parent()
            throws Exception {
        //given
        final ApplicationMenu parentAppMenu1 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "Main1"));
        final ApplicationMenu parentAppMenu2 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "Main2"));

        final ApplicationMenuCreator childCreator1 = new ApplicationMenuCreator(application.getId(), "Child");
        childCreator1.setParentId(parentAppMenu1.getId());
        final ApplicationMenuCreator childCreator2 = new ApplicationMenuCreator(application.getId(), "Child");
        childCreator2.setParentId(parentAppMenu1.getId());
        final ApplicationMenuCreator childCreator3 = new ApplicationMenuCreator(application.getId(), "Child");
        childCreator3.setParentId(parentAppMenu1.getId());

        final ApplicationMenu childCreatedAppMenu1 = getLivingApplicationAPI().createApplicationMenu(childCreator1);
        final ApplicationMenu childCreatedAppMenu2 = getLivingApplicationAPI().createApplicationMenu(childCreator2);
        final ApplicationMenu childCreatedAppMenu3 = getLivingApplicationAPI().createApplicationMenu(childCreator3);

        //when move up
        ApplicationMenu updatedChildMenu = getLivingApplicationAPI().updateApplicationMenu(childCreatedAppMenu3.getId(),
                new ApplicationMenuUpdater().setIndex(1));

        //then
        assertThat(updatedChildMenu.getIndex()).isEqualTo(1);
        assertThat(getLivingApplicationAPI().getApplicationMenu(childCreatedAppMenu1.getId()).getIndex()).isEqualTo(2);
        assertThat(getLivingApplicationAPI().getApplicationMenu(childCreatedAppMenu2.getId()).getIndex()).isEqualTo(3);

        //when move down
        updatedChildMenu = getLivingApplicationAPI().updateApplicationMenu(childCreatedAppMenu3.getId(),
                new ApplicationMenuUpdater().setIndex(2));

        //then
        assertThat(updatedChildMenu.getIndex()).isEqualTo(2);
        assertThat(getLivingApplicationAPI().getApplicationMenu(childCreatedAppMenu1.getId()).getIndex()).isEqualTo(1);
        assertThat(getLivingApplicationAPI().getApplicationMenu(childCreatedAppMenu2.getId()).getIndex()).isEqualTo(3);

        //when change parent
        updatedChildMenu = getLivingApplicationAPI().updateApplicationMenu(childCreatedAppMenu3.getId(),
                new ApplicationMenuUpdater().setParentId(parentAppMenu2.getId()));

        //then
        assertThat(updatedChildMenu.getIndex()).isEqualTo(1);
        assertThat(getLivingApplicationAPI().getApplicationMenu(childCreatedAppMenu1.getId()).getIndex()).isEqualTo(1);
        assertThat(getLivingApplicationAPI().getApplicationMenu(childCreatedAppMenu2.getId()).getIndex()).isEqualTo(2);

    }

    @Test
    public void getApplicationMenu_should_return_the_applicationMenu_identified_by_the_given_id() throws Exception {
        //given
        final ApplicationMenuCreator creator = new ApplicationMenuCreator(application.getId(), "Main", appPage.getId());
        final ApplicationMenu createdAppMenu = getLivingApplicationAPI().createApplicationMenu(creator);

        //when
        final ApplicationMenu retrievedMenu = getLivingApplicationAPI().getApplicationMenu(createdAppMenu.getId());

        //then
        assertThat(retrievedMenu).isNotNull();
        assertThat(retrievedMenu).isEqualTo(createdAppMenu);

    }

    @Test
    public void deleteApplicationMenu_should_remove_the_applicationMenu_identified_by_the_given_id() throws Exception {
        //given
        final ApplicationMenuCreator creator = new ApplicationMenuCreator(application.getId(), "Main", appPage.getId());
        final ApplicationMenu createdAppMenu = getLivingApplicationAPI().createApplicationMenu(creator);

        //when
        getLivingApplicationAPI().deleteApplicationMenu(createdAppMenu.getId());

        //then
        verifyNotExists(createdAppMenu);
    }

    @Test
    public void deleteApplicationMenu_should_remove_sub_menus() throws Exception {
        //given
        Application application = getLivingApplicationAPI()
                .createApplication(new ApplicationCreator("app2", "My secpond app", "1.0"));
        ApplicationPage appPage = getLivingApplicationAPI().createApplicationPage(application.getId(),
                getPage().getId(), "myPage");
        ApplicationMenu mainMenu = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "Main"));
        ApplicationMenu mainMenu2 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "Main2"));
        ApplicationMenu subMenu = getLivingApplicationAPI().createApplicationMenu(
                new ApplicationMenuCreator(application.getId(), "Main", appPage.getId()).setParentId(mainMenu.getId()));

        //when
        getLivingApplicationAPI().deleteApplicationMenu(mainMenu.getId());

        //then
        verifyNotExists(mainMenu);
        verifyNotExists(subMenu);
        verifyExists(mainMenu2);
        verifyExists(appPage);
    }

    @Test
    public void deleteApplication_also_deletes_application_pages_and_applicationMenu() throws Exception {
        //given
        final Application application = getLivingApplicationAPI()
                .createApplication(new ApplicationCreator("app2", "My secpond app", "1.0"));
        final ApplicationPage appPage = getLivingApplicationAPI().createApplicationPage(application.getId(),
                getPage().getId(), "myPage");
        final ApplicationMenu mainMenu = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "Main"));
        ApplicationMenuCreator subMenuCreator = new ApplicationMenuCreator(application.getId(), "Main",
                appPage.getId());
        subMenuCreator.setParentId(mainMenu.getId());
        final ApplicationMenu subMenu = getLivingApplicationAPI().createApplicationMenu(subMenuCreator);

        //when
        getLivingApplicationAPI().deleteApplication(application.getId());

        //then
        verifyNotExists(mainMenu);
        verifyNotExists(subMenu);
        verifyNotExists(appPage);

    }

    @Test
    public void deleteApplicationPage_also_deletes_related_applicationMenu() throws Exception {
        //given
        final ApplicationPage pageToDelete = getLivingApplicationAPI().createApplicationPage(application.getId(),
                getPage().getId(), "pageToDelete");
        final ApplicationPage pageToKeep = getLivingApplicationAPI().createApplicationPage(application.getId(),
                getPage().getId(), "pageToKeep");
        final ApplicationMenu menuToDelete = getLivingApplicationAPI().createApplicationMenu(
                new ApplicationMenuCreator(application.getId(), "Main", pageToDelete.getId()));
        final ApplicationMenu menuToKeep = getLivingApplicationAPI().createApplicationMenu(
                new ApplicationMenuCreator(application.getId(), "Main", pageToKeep.getId()));
        final ApplicationMenu containerMenu = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "Main"));

        //when
        getLivingApplicationAPI().deleteApplicationPage(pageToDelete.getId());

        //then
        // container menu and menu related to another page are keep
        verifyExists(containerMenu);
        verifyExists(menuToKeep);

        // menu related application is deleted
        verifyNotExists(menuToDelete);
    }

    private void verifyExists(ApplicationMenu applicationMenu) throws ApplicationMenuNotFoundException {
        ApplicationMenu retrievedMenu = getLivingApplicationAPI().getApplicationMenu(applicationMenu.getId());
        assertThat(retrievedMenu).isNotNull();
    }

    private void verifyExists(ApplicationPage applicationPage) throws ApplicationPageNotFoundException {
        ApplicationPage retrievedAppPage = getLivingApplicationAPI().getApplicationPage(applicationPage.getId());
        assertThat(retrievedAppPage).isNotNull();
    }

    private void verifyNotExists(ApplicationMenu applicationMenu) {
        try {
            getLivingApplicationAPI().getApplicationMenu(applicationMenu.getId()); //throws exception
            fail("exception expected");
        } catch (ApplicationMenuNotFoundException e) {
            //OK
        }
    }

    private void verifyNotExists(ApplicationPage applicationPage) {
        try {
            getLivingApplicationAPI().getApplicationPage(applicationPage.getId()); //throws exception
            fail("exception expected");
        } catch (ApplicationPageNotFoundException e) {
            //OK
        }
    }

    @Test
    public void searchApplicationMenus_without_filters_without_search_term_should_return_all_applicationMenues_pagged()
            throws Exception {
        //given
        final ApplicationMenu menu1 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "first", appPage.getId()));
        final ApplicationMenu menu2 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second", appPage.getId()));
        final ApplicationMenu menu3 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "third", appPage.getId()));

        //when
        final SearchResult<ApplicationMenu> firstPage = getLivingApplicationAPI()
                .searchApplicationMenus(buildSearchOptions(0, 2));
        final SearchResult<ApplicationMenu> secondPage = getLivingApplicationAPI()
                .searchApplicationMenus(buildSearchOptions(2, 2));

        //then
        assertThat(firstPage).isNotNull();
        assertThat(firstPage.getCount()).isEqualTo(3);
        assertThat(firstPage.getResult()).containsExactly(menu1, menu2);
        assertThat(secondPage).isNotNull();
        assertThat(secondPage.getCount()).isEqualTo(3);
        assertThat(secondPage.getResult()).containsExactly(menu3);
    }

    @Test
    public void searchApplicationMenus_can_filter_on_displayname() throws Exception {
        //given
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "first", appPage.getId()));
        final ApplicationMenu menu2 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second", appPage.getId()));
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "third", appPage.getId()));

        //when
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(0, 10);
        builder.filter(ApplicationMenuSearchDescriptor.DISPLAY_NAME, "second");
        final SearchResult<ApplicationMenu> searchResult = getLivingApplicationAPI()
                .searchApplicationMenus(builder.done());

        //then
        assertThat(searchResult).isNotNull();
        assertThat(searchResult.getCount()).isEqualTo(1);
        assertThat(searchResult.getResult()).containsExactly(menu2);
    }

    @Test
    public void searchApplicationMenus_can_filter_on_index() throws Exception {
        //given
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "first", appPage.getId()));
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second", appPage.getId()));
        final ApplicationMenu menu3 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "third", appPage.getId()));

        //when
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(0, 10);
        builder.filter(ApplicationMenuSearchDescriptor.INDEX, 3);
        final SearchResult<ApplicationMenu> searchResult = getLivingApplicationAPI()
                .searchApplicationMenus(builder.done());

        //then
        assertThat(searchResult).isNotNull();
        assertThat(searchResult.getCount()).isEqualTo(1);
        assertThat(searchResult.getResult()).containsExactly(menu3);
    }

    @Test
    public void searchApplicationMenus_can_filter_on_applicationPageId() throws Exception {
        //given
        final ApplicationPage appPage2 = getLivingApplicationAPI().createApplicationPage(application.getId(),
                getPage().getId(), "mySecondPage");
        final ApplicationMenu menu1 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "first", appPage.getId()));
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second", appPage2.getId()));
        final ApplicationMenu menu3 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "third", appPage.getId()));

        //when
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(0, 10);
        builder.filter(ApplicationMenuSearchDescriptor.APPLICATION_PAGE_ID, appPage.getId());
        final SearchResult<ApplicationMenu> searchResult = getLivingApplicationAPI()
                .searchApplicationMenus(builder.done());

        //then
        assertThat(searchResult).isNotNull();
        assertThat(searchResult.getCount()).isEqualTo(2);
        assertThat(searchResult.getResult()).containsExactly(menu1, menu3);
    }

    @Test
    public void searchApplicationMenus_can_filter_on_applicationId() throws Exception {
        //given
        final Application application2 = getLivingApplicationAPI()
                .createApplication(new ApplicationCreator("app2", "My second app", "1.0"));
        final ApplicationPage appPage2 = getLivingApplicationAPI().createApplicationPage(application2.getId(),
                getPage().getId(), "mySecondPage");

        final ApplicationMenu menu1 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application2.getId(), "first", appPage2.getId()));
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second", appPage.getId()));
        final ApplicationMenu menu3 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application2.getId(), "third", appPage2.getId()));

        //when
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(0, 10);
        builder.filter(ApplicationMenuSearchDescriptor.APPLICATION_ID, application2.getId());
        final SearchResult<ApplicationMenu> searchResult = getLivingApplicationAPI()
                .searchApplicationMenus(builder.done());

        //then
        assertThat(searchResult).isNotNull();
        assertThat(searchResult.getCount()).isEqualTo(2);
        assertThat(searchResult.getResult()).containsExactly(menu1, menu3);
    }

    @Test
    public void searchApplicationMenus_can_filter_on_parentId() throws Exception {
        //given
        final ApplicationMenu parentMenu = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "parent"));

        final ApplicationMenuCreator creator = new ApplicationMenuCreator(application.getId(), "child",
                appPage.getId());
        creator.setParentId(parentMenu.getId());
        final ApplicationMenu childMenu = getLivingApplicationAPI().createApplicationMenu(creator);

        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "third", appPage.getId()));

        //when
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(0, 10);
        builder.filter(ApplicationMenuSearchDescriptor.PARENT_ID, parentMenu.getId());
        final SearchResult<ApplicationMenu> searchResult = getLivingApplicationAPI()
                .searchApplicationMenus(builder.done());

        //then
        assertThat(searchResult).isNotNull();
        assertThat(searchResult.getCount()).isEqualTo(1);
        assertThat(searchResult.getResult()).containsExactly(childMenu);
    }

    @Test
    public void searchApplicationMenus_can_use_searchTerm() throws Exception {
        //given
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "first", appPage.getId()));
        final ApplicationMenu menu2 = getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "second", appPage.getId()));
        getLivingApplicationAPI()
                .createApplicationMenu(new ApplicationMenuCreator(application.getId(), "third", appPage.getId()));

        //when
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(0, 10);
        builder.searchTerm("second");
        final SearchResult<ApplicationMenu> searchResult = getLivingApplicationAPI()
                .searchApplicationMenus(builder.done());

        //then
        assertThat(searchResult).isNotNull();
        assertThat(searchResult.getCount()).isEqualTo(1);
        assertThat(searchResult.getResult()).containsExactly(menu2);
    }

    private SearchOptions buildSearchOptions(final int startIndex, final int maxResults) {
        final SearchOptionsBuilder builder = getApplicationMenuSearchBuilder(startIndex, maxResults);
        final SearchOptions options = builder.done();
        return options;
    }

    private SearchOptionsBuilder getApplicationMenuSearchBuilder(final int startIndex, final int maxResults) {
        final SearchOptionsBuilder builder = new SearchOptionsBuilder(startIndex, maxResults);
        builder.sort(ApplicationMenuSearchDescriptor.INDEX, Order.ASC);
        return builder;
    }

}
